#include <ntddk.h>
#include "ring0.h"

NTSTATUS DriverEntry (IN PDRIVER_OBJECT, IN PUNICODE_STRING);
NTSTATUS ia32cpuAddDevice(IN PDRIVER_OBJECT, IN PDEVICE_OBJECT);
NTSTATUS ia32cpuDispatch (IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS ia32cpuDispatchPnp (IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS ia32cpuStartDevice (IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS ia32cpuDispatchPower (IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS ia32cpuDispatchSystemControl (IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS ia32cpuCompletionRoutine (IN PDEVICE_OBJECT, IN PIRP, IN PVOID);
VOID ia32cpuUnload (IN PDRIVER_OBJECT);

#ifdef ALLOC_PRAGMA
#	pragma alloc_text (INIT, DriverEntry)
#	pragma alloc_text (PAGE, ia32cpuAddDevice)
#	pragma alloc_text (PAGE, ia32cpuDispatchPnp)
#	pragma alloc_text (PAGE, ia32cpuDispatchSystemControl)
#	pragma alloc_text (PAGE, ia32cpuUnload)
#	pragma alloc_text (PAGE, ia32cpuDispatch)
#	pragma alloc_text (PAGE, ia32cpuStartDevice)
#endif

#define IA32CPU_DEVICE_NAME	L"\\Device\\ia32cpu0"
#define DOS_DEVICE_NAME		L"\\DosDevices\\ia32cpu"

#define IA32_TAG 'TROP'

typedef struct _LOCAL_DEVICE_INFO {
	PDEVICE_OBJECT DeviceObject;
	PDEVICE_OBJECT NextLowerDriver;
	BOOLEAN Started;
	BOOLEAN Removed;
	BOOLEAN Filler[2];
	IO_REMOVE_LOCK RemoveLock;
} LOCAL_DEVICE_INFO, *PLOCAL_DEVICE_INFO;

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath)
{
	UNREFERENCED_PARAMETER (RegistryPath);

	DriverObject->MajorFunction[IRP_MJ_CREATE]			= ia32cpuDispatch;
	DriverObject->MajorFunction[IRP_MJ_CLOSE]			= ia32cpuDispatch;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]	= ia32cpuDispatch;
	DriverObject->MajorFunction[IRP_MJ_PNP]				= ia32cpuDispatchPnp;
	DriverObject->MajorFunction[IRP_MJ_POWER]			= ia32cpuDispatchPower;
	DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL]	= ia32cpuDispatchSystemControl;

	DriverObject->DriverUnload							= ia32cpuUnload;
	DriverObject->DriverExtension->AddDevice			= ia32cpuAddDevice;

	return STATUS_SUCCESS;
}

NTSTATUS ia32cpuAddDevice (IN PDRIVER_OBJECT DriverObject, IN PDEVICE_OBJECT PhysicalDeviceObject)
{
	NTSTATUS status = STATUS_SUCCESS;
	PDEVICE_OBJECT deviceObject = NULL;
	PLOCAL_DEVICE_INFO deviceInfo; 
	UNICODE_STRING ntDeviceName;
	UNICODE_STRING win32DeviceName;

	PAGED_CODE();

	RtlInitUnicodeString(&ntDeviceName, IA32CPU_DEVICE_NAME);

	if (!NT_SUCCESS(status = IoCreateDevice(DriverObject, sizeof(LOCAL_DEVICE_INFO), &ntDeviceName, IA32CPU_TYPE, 0, FALSE, &deviceObject)))
		return status;

	RtlInitUnicodeString(&win32DeviceName, DOS_DEVICE_NAME);

	if (!NT_SUCCESS(status = IoCreateSymbolicLink(&win32DeviceName, &ntDeviceName)))
	{
		IoDeleteDevice(deviceObject);
		return status;
	}

	deviceInfo = (PLOCAL_DEVICE_INFO) deviceObject->DeviceExtension;

	if(!(deviceInfo->NextLowerDriver = IoAttachDeviceToDeviceStack(deviceObject, PhysicalDeviceObject))) 
	{
		IoDeleteSymbolicLink(&win32DeviceName);
		IoDeleteDevice(deviceObject);
		return STATUS_NO_SUCH_DEVICE;
	}

	IoInitializeRemoveLock (&deviceInfo->RemoveLock, IA32_TAG, 1, 5);

	deviceObject->Flags |= DO_POWER_PAGABLE;

	deviceInfo->DeviceObject = deviceObject;
	deviceInfo->Removed = FALSE;
	deviceInfo->Started = FALSE;

	deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

	return STATUS_SUCCESS;
}

NTSTATUS ia32cpuDispatchPnp (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpStack;
	NTSTATUS status = STATUS_SUCCESS;
	KEVENT event; 
	UNICODE_STRING win32DeviceName;
	PLOCAL_DEVICE_INFO deviceInfo;

	PAGED_CODE();

	deviceInfo = (PLOCAL_DEVICE_INFO) DeviceObject->DeviceExtension;
	irpStack = IoGetCurrentIrpStackLocation(Irp);

	if (!NT_SUCCESS(status = IoAcquireRemoveLock (&deviceInfo->RemoveLock, Irp))) 
	{
		Irp->IoStatus.Status = status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return status;
	}

	switch (irpStack->MinorFunction) 
	{
	case IRP_MN_START_DEVICE:

		IoCopyCurrentIrpStackLocationToNext(Irp);
		KeInitializeEvent(&event, NotificationEvent, FALSE);

		IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE) ia32cpuCompletionRoutine, &event, TRUE, TRUE, TRUE);

		status = IoCallDriver(deviceInfo->NextLowerDriver, Irp);

		if (STATUS_PENDING == status)
			KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);

		if (NT_SUCCESS(status) && NT_SUCCESS(Irp->IoStatus.Status)) 
			if(NT_SUCCESS(status = ia32cpuStartDevice(DeviceObject, Irp)))
			{
				deviceInfo->Started = TRUE;
				deviceInfo->Removed = FALSE;
			}

		Irp->IoStatus.Status = status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		break;

	case IRP_MN_QUERY_STOP_DEVICE:

		Irp->IoStatus.Status = status = STATUS_UNSUCCESSFUL;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		break;

	case IRP_MN_SURPRISE_REMOVAL:

		deviceInfo->Removed = TRUE;
		deviceInfo->Started = FALSE;

		RtlInitUnicodeString(&win32DeviceName, DOS_DEVICE_NAME);
		IoDeleteSymbolicLink(&win32DeviceName); 

	case IRP_MN_QUERY_REMOVE_DEVICE:

		Irp->IoStatus.Status = STATUS_SUCCESS;
		IoSkipCurrentIrpStackLocation(Irp);
		status = IoCallDriver(deviceInfo->NextLowerDriver, Irp);
		break;

	case IRP_MN_REMOVE_DEVICE:

		if(!deviceInfo->Removed)
		{
			deviceInfo->Removed = TRUE;
			deviceInfo->Started = FALSE;

			RtlInitUnicodeString(&win32DeviceName, DOS_DEVICE_NAME);
			IoDeleteSymbolicLink(&win32DeviceName); 
		} 

		IoReleaseRemoveLockAndWait(&deviceInfo->RemoveLock, Irp);

		Irp->IoStatus.Status = STATUS_SUCCESS;
		IoSkipCurrentIrpStackLocation(Irp);
		status = IoCallDriver(deviceInfo->NextLowerDriver, Irp);

		IoDetachDevice(deviceInfo->NextLowerDriver); 
		IoDeleteDevice(DeviceObject);

		return status;

	case IRP_MN_STOP_DEVICE:
	case IRP_MN_CANCEL_REMOVE_DEVICE: 
	case IRP_MN_CANCEL_STOP_DEVICE: 

		Irp->IoStatus.Status = STATUS_SUCCESS;

	default:

		IoSkipCurrentIrpStackLocation (Irp);
		status = IoCallDriver(deviceInfo->NextLowerDriver, Irp);
		break;
	}

	IoReleaseRemoveLock(&deviceInfo->RemoveLock, Irp); 
	return status;
}

NTSTATUS ia32cpuStartDevice (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PLOCAL_DEVICE_INFO deviceInfo = (PLOCAL_DEVICE_INFO)DeviceObject->DeviceExtension;

	PAGED_CODE();

	if (deviceInfo->Removed)
		return STATUS_DELETE_PENDING;
	else
		return STATUS_SUCCESS;
}

NTSTATUS ia32cpuDispatchPower (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PLOCAL_DEVICE_INFO deviceInfo;

	deviceInfo = (PLOCAL_DEVICE_INFO)DeviceObject->DeviceExtension;

	if (deviceInfo->Removed) 
	{
		PoStartNextPowerIrp(Irp);
		Irp->IoStatus.Status = STATUS_DELETE_PENDING;
		IoCompleteRequest(Irp, IO_NO_INCREMENT );
		return STATUS_DELETE_PENDING;
	}

	PoStartNextPowerIrp(Irp);
	IoSkipCurrentIrpStackLocation(Irp);
	return PoCallDriver(deviceInfo->NextLowerDriver, Irp);
}

NTSTATUS ia32cpuDispatchSystemControl (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PLOCAL_DEVICE_INFO deviceInfo;

	PAGED_CODE();

	deviceInfo = (PLOCAL_DEVICE_INFO) DeviceObject->DeviceExtension;
	IoSkipCurrentIrpStackLocation(Irp);
	return IoCallDriver(deviceInfo->NextLowerDriver, Irp);
}


NTSTATUS ia32cpuDispatch (IN PDEVICE_OBJECT pDO, IN PIRP pIrp)
{
	PLOCAL_DEVICE_INFO pLDI;
	PIO_STACK_LOCATION pIrpStack;
	NTSTATUS Status;

	PULONG pIOBuffer; 

	PAGED_CODE();

	pIrp->IoStatus.Information = 0;
	pLDI = (PLOCAL_DEVICE_INFO)pDO->DeviceExtension;

	if (!NT_SUCCESS(Status = IoAcquireRemoveLock (&pLDI->RemoveLock, pIrp))) 
	{
		pIrp->IoStatus.Status = Status;
		IoCompleteRequest (pIrp, IO_NO_INCREMENT);
		return Status;
	}

	if (!pLDI->Started) 
	{
		pIrp->IoStatus.Status = Status = STATUS_DEVICE_NOT_READY;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT );
		IoReleaseRemoveLock(&pLDI->RemoveLock, pIrp); 
		return Status;
	}

	pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

	switch (pIrpStack->MajorFunction)
	{
	case IRP_MJ_CREATE:

		setup();

	case IRP_MJ_CLOSE:

		Status = STATUS_SUCCESS;
		break;

	case IRP_MJ_DEVICE_CONTROL:

		switch (pIrpStack->Parameters.DeviceIoControl.IoControlCode)
		{
		case IOCTL_IA32CPU_READ_MSR:

			pIOBuffer = (PULONG) pIrp->AssociatedIrp.SystemBuffer;
			pIrp->IoStatus.Information = 0;

			if (pIrpStack->Parameters.DeviceIoControl.InputBufferLength != 4)
				Status = STATUS_INVALID_PARAMETER;
			else
			{
				*((uint64 *)pIOBuffer) = rdmsr(*pIOBuffer);
				pIrp->IoStatus.Information = sizeof(uint64);
				Status = STATUS_SUCCESS;
			}

			break;

		case IOCTL_IA32CPU_WRITE_MSR:

			pIOBuffer = (PULONG) pIrp->AssociatedIrp.SystemBuffer;

			if (pIrpStack->Parameters.DeviceIoControl.InputBufferLength != 12)
				Status = STATUS_INVALID_PARAMETER;
			else
			{
				wrmsr(*pIOBuffer, *((uint64 *)(pIOBuffer + 1)));
				Status = STATUS_SUCCESS;
			}

			break;

		default: 
			Status = STATUS_INVALID_PARAMETER;
		}
		break;

	default: 

		Status = STATUS_NOT_IMPLEMENTED;
		break;
	}

	pIrp->IoStatus.Status = Status;

	IoCompleteRequest(pIrp, IO_NO_INCREMENT );
	IoReleaseRemoveLock(&pLDI->RemoveLock, pIrp); 
	return Status;
}

NTSTATUS ia32cpuCompletionRoutine (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
	PKEVENT event = (PKEVENT)Context;

	UNREFERENCED_PARAMETER(DeviceObject);

	if (Irp->PendingReturned)
		IoMarkIrpPending(Irp);

	KeSetEvent(event, 0, FALSE);

	return STATUS_MORE_PROCESSING_REQUIRED;
}

VOID ia32cpuUnload (IN PDRIVER_OBJECT DriverObject)
{
	PAGED_CODE ();

	return;
}
